theory Redundant_Rules
imports
  "../Rewriting/Trs"
begin

lemma rsteps_union_simulate:
  assumes "\<forall>(l, r) \<in> S. (l, r) \<in> (rstep R)^*"
  shows "(rstep R)^* = (rstep (R \<union> S))^*"
proof
  show "(rstep R)^* \<subseteq> (rstep (R \<union> S))^*"
    using rstep_mono[of R "R \<union> S"] rtrancl_mono by blast
  show "(rstep R)^* \<supseteq> (rstep (R \<union> S))^*"
  proof
    fix s t
    assume "(s, t) \<in> (rstep (R \<union> S))^*"
    then show "(s, t) \<in> (rstep R)^*"
    proof (induct rule: rtrancl_induct)
      case (step u t)
      then obtain l r C \<sigma> where ut: "(u, t) \<in> rstep_r_c_s (l, r) C \<sigma>" and lr: "(l, r) \<in> R \<union> S"
        unfolding rstep_iff_rstep_r_c_s by auto
      from lr have "(u, t) \<in> (rstep R)^*"
      proof
        assume "(l, r) \<in> R"
        with ut have "(u, t) \<in> rstep R" unfolding rstep_iff_rstep_r_c_s by auto
        thus ?thesis by auto
      next
        assume "(l, r) \<in> S"
        with assms have "(l, r) \<in> (rstep R)^*" by auto
        moreover from ut have "u = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>" unfolding rstep_r_c_s_def by auto
        ultimately show ?thesis using rsteps_closed_ctxt rsteps_closed_subst by metis
      qed
      then show ?case using step(3) by auto
    qed auto
  qed
qed

corollary redundant_rules:
  assumes "\<forall>(l, r) \<in> S. (l, r) \<in> (rstep R)^*"
  shows "CR (rstep (R \<union> S)) \<longleftrightarrow> CR (rstep R)"
  unfolding CR_on_def join_def rtrancl_converse
  using rsteps_union_simulate[OF assms] by auto

(* TODO: move *)
lemma conversion_ctxt_closed:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
using assms rsteps_closed_ctxt
unfolding conversion_def
by (simp only: rstep_simps(5)[symmetric])

lemma conversion_subst_closed:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(s \<cdot> \<sigma>,  t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
using assms rsteps_closed_subst
unfolding conversion_def
by (simp only: rstep_simps(5)[symmetric])


lemma rstep_simulate_conv:
  assumes "\<forall>(l, r) \<in> S. (l, r) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(rstep S) \<subseteq> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
proof
  fix s t
  assume "(s, t) \<in> rstep S"
  then obtain l r C \<sigma> where s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t:"t = C\<langle>r \<cdot> \<sigma>\<rangle>" and lr: "(l, r) \<in> S"
    unfolding rstep_iff_rstep_r_c_s rstep_r_c_s_def by auto
  with assms have "(l, r) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*" by auto
  thus "(s, t) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*" using conversion_ctxt_closed conversion_subst_closed s t by metis
qed

lemma symcl_simulate_conv:
  assumes "\<forall>(l, r) \<in> S. (l, r) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(rstep S)\<^sup>\<leftrightarrow> \<subseteq> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
using rstep_simulate_conv[OF assms]
by (auto) (metis converseD conversion_inv reverseTrs subset_iff)

lemma conv_union_simulate:
  assumes "\<forall>(l, r) \<in> S. (l, r) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(rstep (R \<union> S))\<^sup>\<leftrightarrow>\<^sup>* = (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
proof
  show "(rstep (R \<union> S))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_def
  proof
    fix s t
    assume "(s, t) \<in> ((rstep (R \<union> S))\<^sup>\<leftrightarrow>)\<^sup>*"
    then show "(s, t) \<in> ((rstep R)\<^sup>\<leftrightarrow>)\<^sup>*"
    proof (induct rule: rtrancl_induct)
      case (step u t)
      then have "(u, t) \<in> (rstep R)\<^sup>\<leftrightarrow> \<or> (u, t) \<in> (rstep S)\<^sup>\<leftrightarrow>" by auto
      thus ?case
      proof
        assume "(u, t) \<in> (rstep R)\<^sup>\<leftrightarrow>"
        with step show ?thesis using rtrancl_into_rtrancl by metis
      next
        assume "(u, t) \<in> (rstep S)\<^sup>\<leftrightarrow>"
        with symcl_simulate_conv[OF assms] have "(u, t) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*" by auto
        with step show ?thesis by auto
      qed
    qed simp
  qed
next
  show "(rstep R)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (R \<union> S))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_def
    using rstep_union rtrancl_mono sup.cobounded1 symcl_Un
    by metis
qed

(* TODO : move *)
lemma join_mono:
  assumes "r \<subseteq> s"
  shows "r\<^sup>\<down> \<subseteq> s\<^sup>\<down>"
  unfolding join_def
  using assms relcomp_mono rtrancl_mono converse_mono
by metis

lemma redundant_rules_removal:
  assumes "\<forall>(l, r) \<in> S. (l, r) \<in> (rstep R)\<^sup>\<leftrightarrow>\<^sup>*"
  and "CR (rstep R)"
  shows "CR (rstep (R \<union> S))"
  using conv_union_simulate[OF assms(1)]
    CR_imp_conversionIff_join[OF assms(2)]
    join_mono[OF rstep_mono[of R "R \<union> S"]]
  unfolding CR_iff_conversion_imp_join
  by simp

end
