(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Loga
imports "../TermFun"
begin

section \<open>Example - The Binary Logarithm\<close>

text \<open>Compute the quotient of dividing a natural number by two.\<close>
fun half where
  "half 0 = 0"
| "half (Suc 0) = 0"
| "half (Suc (Suc n)) = Suc (half n)"

text \<open>Compute the binary logarithm by successive bisection.\<close>
function log where
  "log n = (case half n of 0 \<Rightarrow> 0 | Suc m \<Rightarrow> Suc (log (Suc m)))"
  by pat_completeness auto

text \<open>Generate first-order term rewrite systems for the defining equations
of @{const "half"} and @{const "log"}. The @{text "no_sim"} attribute in the
latter case is needed to prevent the system from proving the \emph{simulation lemma}
for @{const "log"} (which is only possible after termination has been established).
(Consult the paper for more details.)\<close>
define_trs half.simps
define_trs (no_sim) log.psimps

text \<open>The resulting TRSs are available under the names @{const "half.trs"} and
@{const "log.trs"}, having the definitions @{thm[source] half.trs_def} and
@{thm[source] log.trs_def}, respectively.\<close>
term half.trs
term log.trs
thm half.trs_def log.trs_def

text \<open>
Finally, we prove termination of @{const "log"}. The initial obligation is
to show that all inputs are in the domain of @{const "log"} (and hence, the function
is defined on all inputs).\<close>
termination log
  txt \<open>
  We first reduce this obligation to termination of @{const "log"}s TRS
  representation (the main content of our paper).
\<close>
  apply (trs_reduce log)
  txt \<open>Before we export the TRS, we transform it into a finite list of explicit rules.\<close>
  apply expand_trs_problem
  txt \<open>The result can directly be written to an external file (as XML).\<close>
  export_termination_problem "log.xml"
  apply (ceta_import_proof "proofs/log.proof.xml")
  apply eval
done

end

